/********************************************************************/
/**                                                                **/
/**   exapmle_2.c                                                  **/
/**                                                                **/
/**   A universal generator for multivariate T-concave             **/
/**   distributions                                                **/
/**                                                                **/
/**   generate normal distributed variates                         **/
/**                                                                **/
/**   domain:  IR^N                                                **/
/**   density: f(x_1,...x_N) = exp(-x_1^2-...-x_N^2)               **/
/**                                                                **/
/**   OVERWRITE DEFAULT VALUES                                     **/
/**                                                                **/
/**   authors: Josef Leydold                                       **/
/**   last modified: Mon Mar  9 13:38:00 MET 1998                  **/
/**                                                                **/
/********************************************************************/              

/* the header file */
#include "tdrmv.h"

/*------------------------------------------------------------------*/
/* Compiling options in config.h                                    */
/*                                                                  */
/*    N          2, ..., 14                                         */
/*    MODE       1                                                  */
/*    RECTANGLE  0                                                  */
/*    GAMMA      1 or 2                                             */
/*                                                                  */
/*------------------------------------------------------------------*/

#define SAMPLE_SIZE   1000     /* sample size                       */

/*------------------------------------------------------------------*/

/* print random point */
static void print_rpoint(double *v, int n);

/*------------------------------------------------------------------*/

/*******************************************************************

  Main

  ******************************************************************/

int main()
{
  C_ROOT *hat;         /* pointer to hat */
  C_CONSTR *hc;        /* parameters for construting hat functions */
  double rpoint[N];    /* array for random point */
  int i;

  /* density */
  double initialize_normal(void);  /* initialize density function */
  double normal();                 /* density function */
  void grad_normal();              /* gradient of density */
  
  /*--------------------------------------------------------------*/

  /* default parameters for constructing hat function */
  hc = get_constr_default();

  /*--------------------------------------------------------------*/

  /** overwrite default parameters                               **/
  /* (the default parameters are set in config.h)                 */

  /* volume below density (default is 0. == unknown) */
  hc->volume_density = initialize_normal();

  /* control generation of cones */
  hc->max_cones = MAX_N_CONES;                /* maximum number of cones (at least 2^(N+T_STEPS_MIN) */
  hc->steps_min = T_STEPS_MIN;                /* minimum number of triangulation steps */
  hc->step_tp = OPTIMAL_TP_STEP;              /* triangulation step when optimal touching points is calculated */

  /* reject bad touching points */
  hc->reject_tp_abs = REJECT_TP_ABS;          /* reject if Hi is greaten than this value */
  hc->reject_tp_rel = REJECT_TP_REL;          /* reject if Hi > median + reject_tp_rel * (percentile - median) */
  hc->reject_tp_percentile = REJECT_TP_PERCENTILE;/* percentile for rejection */
  /* there are no checks if the corresponding parameter is < 0. */

  /* control allocation of memory */
  hc->cone_alloc_block = CONE_ALLOC_BLOCK;    /* number of cones to be allocated at once */
  hc->vertex_alloc_block = VERTEX_ALLOC_BLOCK;/* number of vertices to be allocated at once */

  /* parameters for finding touching points */
  hc->find_tp_start = FIND_TP_START;          /* starting point for finding optimal touching point */
  hc->find_tp_steps_r = FIND_TP_STEPS_R;      /* max number of steps for finding proper touching point, --> Infinity */
  hc->find_tp_steps_l = FIND_TP_STEPS_L;      /* max number of steps for finding proper touching point, --> 0 */
  hc->find_tp_step_size = FIND_TP_STEP_SIZE;  /* size of each step for finding proper touching point */
  hc->find_tp_steps_lb = FIND_TP_STEPS_LB;    /* max number of steps for finding lower bound of bracket */
  hc->find_tp_steps_ub = FIND_TP_STEPS_UB;    /* max number of steps for finding upper bound of bracket */
  hc->find_tp_tol = FIND_TP_TOL;              /* acceptable tolerance for Brent's algorithm for finding minimum */
  hc->find_tp_steps_brent = FIND_TP_STEPS_BRENT;/* max number of interation steps for Brent's algorithm */

  /* parameters for finding mode */
  /* see hooke.c for explanation */
  hc->hooke_rho_begin = HOOKE_RHO_BEGIN;      /* stepsize geometric shrink */
  hc->hooke_rho_second = HOOKE_RHO_SECOND;    /* stepsize geometric shrink (second run, only if > 0) */
  hc->hooke_epsilon = HOOKE_EPSILON;          /* ending value of stepsize  */
  hc->hooke_imax = HOOKE_IMAX;                /* max # of iterations       */
  hc->mode_to_boundary = MODE_TO_BOUNDARY;    /* move mode to boundary if |mode - boundary| / length < MODE_TO_BOUNDARY */

  /* size of search table */
  hc->guide_rel_size = GUIDE_TABLE_SIZE;      /* relative size of guide table compared to number of cones */

  /*--------------------------------------------------------------*/

  /* hat */
  hat = get_hat(hc,normal,grad_normal);

  /*--------------------------------------------------------------*/
  
  /* sample */
  for( i=0; i<SAMPLE_SIZE; i++ ) {
    sample_point(hat,rpoint);
    print_rpoint(rpoint,N);
  }

  /*--------------------------------------------------------------*/
      
#if DEBUG > 0
  db_close(hat);
#endif
  
  exit(0);
} /* end of main() */

/*-----------------------------------------------------------------*/

static void print_rpoint(double *v, int n)
/*
  print random point
*/
{
  int i;

  printf("( %f",v[0]);
  for( i=1; i<n; i++ )
    printf(", %f",v[i]);
  printf(" )\n");

} /* end of db_print_rpoint() */

/*-----------------------------------------------------------------*/

/*******************************************************************

  standard normal distribution

  ******************************************************************/

/*-----------------------------------------------------------------*/

double initialize_normal(void)
/*
  returns volume below density function
*/
{
  int i;
  double volume_density;   /* volume below density */
  
  /* compute volume below density */
  volume_density = 1.;
  for( i=0; i<N; i++ )
    volume_density *= PI;
  return( sqrt(volume_density) );

} /* end of initialize_normal() */

/*-----------------------------------------------------------------*/

double normal(double *coord)
/*
  normal density function
*/
{
  int i;             /* aux counter */
  double f=0;       /* density function */

  /* gaussian density */
  for( i=0; i<N; i++ )
    f += coord[i] * coord[i];

  return( exp(-f) );
} /* end of normal() */

/*-----------------------------------------------------------------*/

void grad_normal(double *coord, double *grad)
/*
  gradient of normal density function
*/
{
  int i;              /* aux counter */
  double f=0;             /* density at point */

  /* d(e^{-\sum j*x_j^2})/d x_i = -2*i*x_i*e^{-\sum j*x_j^2} */

  for( i=0; i<N; i++ )
    f += coord[i] * coord[i];
  f = exp( -f );

  for( i=0; i<N; i++ )
    grad[i] = (-2) * coord[i] * f;

} /* end of grad_normal() */

/*-----------------------------------------------------------------*/

