/********************************************************************/
/**                                                                **/
/**   exapmle_3.c                                                  **/
/**                                                                **/
/**   A universal generator for multivariate T-concave             **/
/**   distributions                                                **/
/**                                                                **/
/**   generate normal distributed variates                         **/
/**                                                                **/
/**   domain:  rectangle                                           **/
/**   mode:    (m_1,...m_N)    (see below)                         **/
/**   density: f(x_1,...x_N) = exp(-(x_1-m_1)^2-...-(x_N-m_N)^2)   **/
/**                                                                **/
/**                                                                **/
/**   authors: Josef Leydold                                       **/
/**   last modified: Mon Mar  9 13:38:00 MET 1998                  **/
/**                                                                **/
/********************************************************************/              

/* the header file */
#include "tdrmv.h"

/*------------------------------------------------------------------*/
/* Compiling options in config.h                                    */
/*                                                                  */
/*    N          2, ..., 14                                         */
/*    MODE       1                                                  */
/*    RECTANGLE  1                                                  */
/*    GAMMA      2                                                  */
/*                                                                  */
/*------------------------------------------------------------------*/

#define SAMPLE_SIZE   1000      /* sample size                       */

/* mode of unrestricted density function */
static const double mode[] = {2.1,1.5,2.,1.,1.,1.,1.,1.,1.,1.,1.,1.,1.,1.};
/* static const double mode[] = {0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}; */

/* bounds of rectangle */
//static const double lower[] = {-1.,-1.,-1.,-1.,-1.,-1.,-1.,-1.,-1.,-1.,-1.,-1.,-1.,-1.};
static const double lower[] = {1.,1.,1.,1.,-1.,-1.,-1.,-1.,-1.,-1.,-1.,-1.,-1.,-1.};
static const double upper[] = {2.,2.,2.,2.,2.,2.,2.,2.,2.,2.,2.,2.,2.,2.};

/*------------------------------------------------------------------*/

/* print random point */
static void print_rpoint(double *v, int n);

/*------------------------------------------------------------------*/

/*******************************************************************

  Main

  ******************************************************************/

int main()
{
  C_ROOT *hat;         /* pointer to hat */
  C_CONSTR *hc;        /* parameters for construting hat functions */
  double rpoint[N];    /* array for random point */
  int i;

  /* density */
  double initialize_normal();      /* initialize density function */
  double normal();                 /* density function */
  void grad_normal();              /* gradient of density */
  
  /*--------------------------------------------------------------*/

  /* default parameters for constructing hat function */
  hc = get_constr_default();

  /*--------------------------------------------------------------*/

  /** overwrite default parameters                               **/
  /* (the default parameters are set in config.h)                 */

  /* control generation of cones */
  hc->steps_min = 4;                          /* minimum number of triangulation steps */
  hc->step_tp = OPTIMAL_TP_STEP;              /* triangulation step when optimal touching points is calculated */

  /* parameters for finding touching points */
  hc->find_tp_start = 1.;                     /* starting point for finding optimal touching point */
  hc->find_tp_steps_r = 10;                   /* max number of steps for finding proper touching point, --> Infinity */
  hc->find_tp_steps_l = 10;                   /* max number of steps for finding proper touching point, --> 0 */

  /*--------------------------------------------------------------*/

#if RECTANGLE == 1
  /* set rectangle */
  for( i=0; i<N; i++ ) {
    hc->rl[i] = lower[i];
    hc->ru[i] = upper[i];
  }
#endif

  /* volume below density (default is 0. == unknown) */
  hc->volume_density = initialize_normal(hc);

  /*--------------------------------------------------------------*/

#if MODE == 1
  /* get mode of density                             */
  /*                                                 */
  /* the mode is stored in hc->mode[N]               */
  /* default hc->mode = ( 0, ..., 0 )                */
  /* the function get_mode search for the mode,      */
  /* using hc->mode as its starting point            */

  /* mode / starting point for searching mode */
  for( i=0; i<N; i++ )
    hc->mode[i] = 0.5;

  /* find mode */
  get_mode(hc,normal);
#endif  

  /*--------------------------------------------------------------*/

  /* hat */
  hat = get_hat(hc,normal,grad_normal);

  /*--------------------------------------------------------------*/

  /* sample */
  for( i=0; i<SAMPLE_SIZE; i++ ) {
    sample_point(hat,rpoint);
    print_rpoint(rpoint,N);
  }

  /*--------------------------------------------------------------*/

#if DEBUG > 0
  db_close(hat);
#endif
  
  exit(0);
} /* end of main() */

/*-----------------------------------------------------------------*/
static void print_rpoint(double *v, int n)
/*
  print random point
*/
{
  int i;

  printf("( %f",v[0]);
  for( i=1; i<n; i++ )
    printf(", %f",v[i]);
  printf(" )\n");

} /* end of db_print_rpoint() */

/*-----------------------------------------------------------------*/


/*******************************************************************

  standard normal distribution

  ******************************************************************/

/*-----------------------------------------------------------------*/

double initialize_normal(C_CONSTR *hc)
/*
  returns volume below density function
*/
{
  int i;
  double volume_density;   /* volume below density */

  /* compute volume below density */
#if RECTANGLE == 1
  volume_density = 1.;
  for( i=0; i<N; i++ )
    volume_density *= sqrt(PI) * 0.5 * ( erf( hc->ru[i] - mode[i] ) - erf( hc->rl[i] - mode[i] ) );
  return( volume_density );
#else
  volume_density = 1.;
  for( i=0; i<N; i++ )
    volume_density *= PI;
  return( sqrt(volume_density) );
#endif

} /* end of initialize_density() */

/*-----------------------------------------------------------------*/

double normal(double *coord)
/*
  normal density function
*/
{
  int i;             /* aux counter */
  double f=0;       /* density function */

  /* gaussian density */
  for( i=0; i<N; i++ )
    f += ( coord[i] - mode[i] ) * ( coord[i] - mode[i] );

  return( exp(-f) );
} /* end of normal() */

/*-----------------------------------------------------------------*/

void grad_normal(double *coord, double *grad)
/*
  gradient of normal density function
*/
{
  int i;              /* aux counter */
  double f=0;             /* density at point */

  /* d(e^{-\sum j*x_j^2})/d x_i = -2*i*x_i*e^{-\sum j*x_j^2} */

  for( i=0; i<N; i++ )
    f += ( coord[i] - mode[i] ) * ( coord[i] - mode[i] );
  f = exp( -f );

  for( i=0; i<N; i++ )
    grad[i] = (-2.) * ( coord[i] - mode[i] ) * f;

} /* end of grad_normal() */

/*-----------------------------------------------------------------*/

