/********************************************************************/
/**                                                                **/
/**   dgs.c                                                        **/
/**                                                                **/
/**   A universal generator for multivariate T-concave             **/
/**   distributions                                                **/
/**                                                                **/
/**   generator for gamma variates                                 **/
/**                                                                **/
/**   authors: see below                                           **/
/**   minor changes by Josef Leydold                               **/
/**   last modified: Mon Mar  9 13:38:00 MET 1998                  **/
/**                                                                **/
/********************************************************************/              

/* the header file */
#include "tdrmv.h"

/*------------------------------------------------------------------*/
#if GAMMA == 1
/*------------------------------------------------------------------*/

/*********************************************************************

  gds()

  *******************************************************************/

/* functions */
double gds(double parameter);   /* generator for gamma distribution */
double nsc(void);               /* generator normal distribution    */

extern double uniform();        /* uniform random number generator  */

/*------------------------------------------------------------------*/

/***************************************************************
 * Gamma Distribution - Rejection algorithm gs combined with   *
 *                      Acceptance complement method gd        *
 ***************************************************************
 * FUNCTION: - gds samples a random number from the gamma      *
 *                 distribution with shape parameter a > 0     *
 *                 Acceptance rejection gs for a < 1           *
 *                 Acceptance complement gd for a >= 1         *
 * REFERENCES: -   J.H. Ahrens, U. Dieter 1974:                *
 *                 Computer Methods For Sampling From Gamma,   *
 *                 Beta, Poisson and Binomial Distributions.   *
 *                 Computing 12, pp. 223 - 246.                *
 *             -   J.H. Ahrnes, U. Dieter 1982: Generating     *
 *                 Gamma By A Modified Rejection Technique.    *
 *                 Communications of the ACM 25, pp. 47 - 54.  *
 * SUBPROGRAMS:-   uniform()  .. (0,1)-uniform generator       *
 *               - NORMAL() ... normal generator N(0,1)        *
 *                                                             *
 * Implemented by G. Thallinger and E. Stadlober, August 1990  *
 ***************************************************************/
double gds(parameter)
double  parameter;
{
  double a=parameter;                 /* a...scale parameter */

static double aa = -1.0, aaa = -1.0,
       b=0., c=0., d=0., e=0., r=0., s=0., si=0., ss=0., q0=0.,
       q1 = 0.0416666664, q2 =  0.0208333723, q3 = 0.0079849875,
       q4 = 0.0015746717, q5 = -0.0003349403, q6 = 0.0003340332,
       q7 = 0.0006053049, q8 = -0.0004701849, q9 = 0.0001710320,
       a1 = 0.333333333,  a2 = -0.249999949,  a3 = 0.199999867,
       a4= -0.166677482,  a5 =  0.142873973,  a6= -0.124385581,
       a7 = 0.110368310,  a8 = -0.112750886,  a9 = 0.104089866,
       e1 = 1.000000000,  e2 =  0.499999994,  e3 = 0.166666848,
       e4 = 0.041664508,  e5 =  0.008345522,  e6 = 0.001353826,
       e7 = 0.000247453;

double gds,p,q,t,sign_u,u,v,w,x;

if (a < 1.0)
  {               /* CASE A: Acceptance rejection algorithm gs */
   b = 1.0 + 0.36788794412 * a;                      /* Step 1 */

   for(;;)
      {p = b *  uniform();
       if (p <= 1.0)
         {gds = exp(log(p) / a);      /* Step 2. Case gds <= 1 */
          if (log( uniform()) <= -gds) return(gds);
         }
       else
         {gds = - log ((b - p) / a);   /* Step 3. Case gds > 1 */
          if (log( uniform()) <= ((a - 1.0) * log(gds)))
             return(gds);
         }
      }
  }
else
  {if (a != aa)  /* CASE B: Acceptance complement algorithm gd */
     {aa = a;                          /* Step 1. Preparations */
      ss = a - 0.5;
      s = sqrt(ss);
      d = 5.656854249 - 12.0 * s;
     }

   t = nsc();                  /* Step 2. Normal deviate */
                               
   x = s + 0.5 * t; 
   gds = x * x;
   if (t >= 0.0)
      return(gds);                     /* Immediate acceptance */

   u =  uniform();          /* Step 3. Uniform random number */
   if (d * u <= t * t * t) return(gds);  /* Squeeze acceptance */

   if (a != aaa)
     {aaa = a;                  /* Step 4. Set-up for hat case */
      r = 1.0 / a;
      q0 = ((((((((q9 * r + q8) * r + q7) * r + q6) *  r + q5) *
           r + q4) * r + q3) * r + q2) * r + q1) * r;
      if (a > 3.686)
        {if (a > 13.022)
           {b = 1.77;
            si = 0.75;
            c = 0.1515 / s;
           }
         else
           {b = 1.654 + 0.0076 * ss;
            si = 1.68 / s + 0.275;
            c = 0.062 / s + 0.024;
           }
        }

      else
        {b = 0.463 + s - 0.178 * ss;
         si = 1.235;
         c = 0.195 / s - 0.079 + 0.016 * s;
        }
      }

    if (x > 0.0)                   /* Step 5. Calculation of q */
      {v = t / (s + s);                             /* Step 6. */
       if (fabs(v) > 0.25)
         {q = q0 - s * t + 0.25 * t * t + (ss+ss) * log(1.0 + v);
         }
       else
         {q = q0 + 0.5 * t * t * ((((((((a9 * v + a8) *
              v + a7) * v + a6) *  v + a5) * v + a4) *
              v + a3) * v + a2) * v + a1) * v;
         }                      /* Step 7. Quotient acceptance */
       if (log(1.0 - u) <= q)
          return(gds);
      }

    for(;;)
      {                /* Step 8. Double exponential deviate t */
       do
         {e = -log( uniform());
          u =  uniform();
          u = u + u - 1.0;
          sign_u = (u > 0)? 1.0 : -1.0;
          t = b + (e * si) * sign_u;
         }                           /* Step 9. Rejection of t */
       while (t <= -0.71874483771719);

    v = t / (s + s);                      /* Step 10. New q(t) */
    if (fabs(v) > 0.25)
      {q = q0 - s * t + 0.25 * t * t + (ss + ss) * log(1.0 + v);
      }
    else
      {q = q0 + 0.5 * t * t * ((((((((a9 * v + a8) * v + a7) *
           v + a6) * v + a5) * v + a4) * v + a3) * v + a2) *
           v + a1) * v;
      }
    if (q <= 0.0) continue;                        /* Step 11. */
    if (q > 0.5)
      {w = exp(q) - 1.0;
      }
    else
      {w = ((((((e7 * q + e6) * q + e5) * q + e4) * q + e3) *
           q + e2) * q + e1) * q;
      }                             /* Step 12. Hat acceptance */
    if ( c * u * sign_u <= w * exp(e - 0.5 * t * t))
      {x = s + 0.5 * t;
       return(x*x);
      }
    }
  }
}

/*------------------------------------------------------------------*/

/***************************************************************
 * NORMAL Distribution - NSC (Sinus-Cosinus or Box/Muller      *
 *                       Method)                               *
 ***************************************************************
 * FUNCTION: -    nsc samples a random number from the         *
 *                standard normal distribution                 *
 * REFERENCE: -   G.E.P. Box and M.E. Muller: A note on the    *
 *                generation of random normal deviates.        *
 *                Annals of Mathem. Stat. 29, (1958), 610-611. *
 * SUBPROGRAMS: - uniform()   ... (0,1)--uniform generator,    *
 *                contained in cranduni.c                      *
 ***************************************************************/
double nsc(void)
{
  double u,v,s;
  static double x2;
  static double two_pi = 2.0 * M_PI;
  static char f = 1;

  f = -f;
  if (f > 0) return(x2);
  u =  uniform();
  v =  uniform();
  s = sqrt(-2.0 * log(u));
  x2 = s * sin(two_pi * v);
  return(s * cos(two_pi * v));
}

/*------------------------------------------------------------------*/
#endif
/*------------------------------------------------------------------*/
